try:
    basestring = basestring
except NameError:
    # Python 3
    basestring = str

try:
    from cStringIO import StringIO
    BytesIO = StringIO
except ImportError:
    # Python 3
    from io import StringIO, BytesIO  # noqa

try:
    from html import escape
except ImportError:
    # Python 2
    from cgi import escape  # noqa



import unittest
import transaction
import gc
import json

from ZODB.interfaces import IDatabase
from ZODB.utils import u64, p64, tid_repr, oid_repr
from persistent import Persistent
from zope.app.container.btree import BTreeContainer
from zope.app.container.interfaces import IContained
from zope.app.testing import setup
from zope.component import provideAdapter, getGlobalSiteManager
from zope.exceptions.interfaces import UserError
from zope.interface import implementer
from zope.publisher.browser import TestRequest
from zope.traversing.interfaces import IContainmentRoot
from zope.security.proxy import Proxy
from zope.security.checker import ProxyFactory

from zodbbrowser.browser import (
    ZodbObjectAttribute, VeryCarefulView, ZodbInfoView, ZodbHistoryView,
    getObjectType, getObjectTypeShort, getObjectPath)
from zodbbrowser.btreesupport import EmptyOOBTreeState
from zodbbrowser.history import ZodbObjectHistory, ZodbHistory, getIterableStorage
from zodbbrowser.state import GenericState, ZodbObjectState
from zodbbrowser.testing import SimpleValueRenderer

from .realdb import RealDatabaseTest


@implementer(IDatabase)
class DatabaseStub(object):

    opened = 0

    def open(self):
        self.opened += 1
        return ConnectionStub(self)


class ConnectionStub(object):
    def __init__(self, db):
        self.db = db

    def close(self):
        self.db.opened -= 1


@implementer(IContainmentRoot)
class RootFolderStub(BTreeContainer):
    pass


class PersistentStub(BTreeContainer):
    pass


@implementer(IContained)
class NonpersistentStub(dict):

    __name__ = __parent__ = None


class TestZodbObjectAttribute(unittest.TestCase):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(SimpleValueRenderer)
        self.attribute = ZodbObjectAttribute('foo', 42, 't565')

    def tearDown(self):
        setup.placelessTearDown()

    def test_rendered_name(self):
        self.assertEqual(self.attribute.rendered_name(),
                         "'foo' [tid=t565]")

    def test_rendered_value(self):
        self.assertEqual(self.attribute.rendered_value(),
                         "42 [tid=t565]")

    def test_repr(self):
        self.assertEqual(repr(self.attribute),
                         "ZodbObjectAttribute('foo', 42, 't565')")

    def test_equality(self):
        self.assertEqual(self.attribute,
                         ZodbObjectAttribute('foo', 42, 't565'))

    def test_inequality(self):
        self.assertNotEqual(self.attribute,
                            ZodbObjectAttribute('foo', 42, 't575'))
        self.assertNotEqual(self.attribute,
                            ZodbObjectAttribute('foo', 43, 't565'))
        self.assertNotEqual(self.attribute,
                            ZodbObjectAttribute('fox', 42, 't565'))
        self.assertNotEqual(self.attribute,
                            object())

    def test_not_equals(self):
        self.assertFalse(self.attribute !=
                         ZodbObjectAttribute('foo', 42, 't565'))
        self.assertTrue(self.attribute != object())


class RandomThing(Persistent):
    pass


class SampleCarefulView(VeryCarefulView):
    def render(self):
        self.context.attr = 42
        transaction.get().join(CustomResource())
        return 'hi'


class CustomResource(object):
    def __repr__(self):
        return '<CustomResource>'

    def abort(self, txn):
        pass


class TestVeryCarefulView(RealDatabaseTest):

    def test_call_undoes_changes(self):
        self.root = self.conn.root()
        self.root['obj'] = obj = RandomThing()
        transaction.commit()
        view = SampleCarefulView(obj, None)
        view()
        self.assertFalse(hasattr(obj, 'attr'))


class TestZodbInfoViewWithRealDb(RealDatabaseTest):

    def _zodbInfoView(self, obj, request):
        view = ZodbInfoView(obj, request)
        view.template = lambda: ''
        view()
        return view

    def setUp(self):
        RealDatabaseTest.setUp(self)
        self.root = self.conn.root()
        self.root['stub'] = PersistentStub()
        self.root['stub']['member'] = NonpersistentStub()
        self.root['stub']['member']['notpersistent'] = 'string'
        self.conn.add(self.root['stub']) # force oid allocation
        self.root['root'] = RootFolderStub()
        self.root['root']['item'] = PersistentStub()
        transaction.commit()
        provideAdapter(GenericState)
        provideAdapter(ZodbObjectHistory)

    def testCall(self):
        request = TestRequest()
        view = self._zodbInfoView(self.root, request)
        self.assertEqual(view(), '')
        self.assertEqual(view.latest, True)

        tid = ZodbObjectState(self.root).tid
        request = TestRequest(form={'tid': tid_repr(tid)})
        view = self._zodbInfoView(self.root, request)
        self.assertEqual(view.latest, False)

        oid = self.root._p_oid
        request = TestRequest(form={'oid': oid_repr(oid)})
        request.annotations['ZODB.interfaces.IConnection'] = self.root._p_jar
        view = self._zodbInfoView(None, request)

    def testGetJar(self):
        view = ZodbInfoView(self.root, TestRequest())
        self.assertEqual(view.jar, self.root._p_jar)
        view = ZodbInfoView(self.root['stub']['member'], TestRequest())
        self.assertEqual(view.jar, self.root._p_jar)

    def testSelectObjectToView_use_context(self):
        view = ZodbInfoView(self.root, TestRequest())
        self.assertEqual(view.selectObjectToView(), self.root)
        view = ZodbInfoView(self.root['root']['item'], TestRequest())
        self.assertEqual(view.selectObjectToView(), self.root['root']['item'])

    def testSelectObjectToView_find_parent(self):
        view = ZodbInfoView(self.root['stub']['member'], TestRequest())
        self.assertEqual(view.selectObjectToView(), self.root['stub'])

    def testSelectObjectToView_find_parent_fail(self):
        view = ZodbInfoView(self.root['stub']['member']['notpersistent'], TestRequest())
        self.assertRaises(Exception, view.selectObjectToView)

    def testSelectObjectToView_find_parent_fail_fall_back_to_root(self):
        view = ZodbInfoView(self.root['stub']['member']['notpersistent'], TestRequest())
        view.jar = self.root._p_jar
        self.assertEqual(view.selectObjectToView(), self.root)

    def testSelectObjectToView_by_oid(self):
        oid = u64(self.root['stub']._p_oid)
        view = ZodbInfoView(self.root, TestRequest(form={'oid': str(oid)}))
        self.assertEqual(view.selectObjectToView(), self.root['stub'])

    def testSelectObjectToView_by_oid_in_hex(self):
        oid = u64(self.root['stub']._p_oid)
        hex_oid = hex(oid).rstrip('L')
        view = ZodbInfoView(self.root, TestRequest(form={'oid': hex_oid}))
        self.assertEqual(view.selectObjectToView(), self.root['stub'])

    def testSelectObjectToView_by_oid_bad_format(self):
        view = ZodbInfoView(self.root, TestRequest(form={'oid': 'dunno'}))
        self.assertRaises(UserError, view.selectObjectToView)

    def testSelectObjectToView_by_oid_bad_value(self):
        view = ZodbInfoView(self.root, TestRequest(form={'oid': '0xdeafbeef'}))
        self.assertRaises(UserError, view.selectObjectToView)

    def testFindClosestPersistent(self):
        view = ZodbInfoView(self.root['stub']['member'], TestRequest())
        self.assertEqual(view.findClosestPersistent(), self.root['stub'])
        view = ZodbInfoView(self.root['stub']['member']['notpersistent'],
                            TestRequest())
        self.assertEqual(view.findClosestPersistent(),
                         None)

    def testGetRequestedTid(self):
        view = ZodbInfoView(self.root, TestRequest())
        self.assertEqual(view.getRequestedTid(), None)
        self.assertEqual(view.getRequestedTidNice(), None)
        view = ZodbInfoView(self.root,
                            TestRequest(form={'tid': '12345678912345678'}))
        self.assertEqual(view.getRequestedTid(), '12345678912345678')
        self.assertEqual(view.getRequestedTidNice(),
                         '1905-05-13 03:32:22.050327')

    def testPrimitiveMethods(self):
        view = self._zodbInfoView(self.root, TestRequest())
        self.assertEqual(view.getObjectId(), u64(self.root._p_oid))
        self.assertTrue('PersistentMapping' in view.getObjectType())
        self.assertEqual(view.getStateTid(),
                         u64(ZodbObjectState(self.root).tid))
        self.assertEqual(view.getStateTidNice(),
                         view._tidToTimestamp(ZodbObjectState(self.root).tid))

    def testLocate(self):
        view = self._zodbInfoView(self.root, TestRequest())
        jsonResult = json.loads(view.locate_json('/'))
        self.assertEqual(jsonResult,
                         {"url": "@@zodbbrowser?oid=0x0",
                          "oid": 0})
        jsonResult = json.loads(view.locate_json('/stub/member/notpersistent'))
        self.assertTrue("partial_url" in jsonResult)
        self.assertTrue("partial_oid" in jsonResult)
        self.assertEqual(jsonResult["error"],
                         "Not persistent: /stub/member/notpersistent")
        jsonResult = json.loads(view.locate_json('/stub/nonexistent'))
        self.assertEqual(jsonResult,
                         {"partial_url": "@@zodbbrowser?oid=0x1",
                          "partial_path": "/stub",
                          "error": "Not found: /stub/nonexistent",
                          "partial_oid": 1})

    def testLocateNoLeadingSlash(self):
        view = self._zodbInfoView(self.root, TestRequest())
        jsonResult = json.loads(view.locate_json('stub'))
        self.assertEqual(jsonResult,
                         {"url": "@@zodbbrowser?oid=0x1",
                          "oid": 1})

    def testLocateStartWithOID(self):
        view = self._zodbInfoView(self.root, TestRequest())
        jsonResult = json.loads(view.locate_json('0'))
        self.assertEqual(jsonResult,
                         {"url": "@@zodbbrowser?oid=0x0",
                          "oid": 0})
        jsonResult = json.loads(view.locate_json('0x0'))
        self.assertEqual(jsonResult,
                         {"url": "@@zodbbrowser?oid=0x0",
                          "oid": 0})
        jsonResult = json.loads(view.locate_json('0x1/nonexistent'))
        self.assertEqual(jsonResult,
                         {"partial_url": "@@zodbbrowser?oid=0x1",
                          "partial_path": "0x1",
                          "error": "Not found: 0x1/nonexistent",
                          "partial_oid": 1})

    def testLocateStartWithOID_that_does_not_exist(self):
        view = self._zodbInfoView(self.root, TestRequest())
        jsonResult = json.loads(view.locate_json('0x1234'))
        self.assertEqual(jsonResult,
                         {"partial_url": "@@zodbbrowser?oid=0x0",
                          "partial_path": "/",
                          "error": "Not found: 0x1234",
                          "partial_oid": 0})

    def testGetUrl(self):
        view = self._zodbInfoView(self.root, TestRequest())
        self.assertEqual(view.getUrl(), '@@zodbbrowser?oid=0x%x' %
                         u64(self.root._p_oid))
        view = self._zodbInfoView(self.root, TestRequest())
        self.assertEqual(view.getUrl(1, 2), '@@zodbbrowser?oid=0x1&tid=0x2')
        view = ZodbInfoView(self.root, TestRequest(form={'tid': '2'}))
        self.assertEqual(view.getUrl(1), '@@zodbbrowser?oid=0x1&tid=2')


class ZodbObjectStateStub(object):

    def __init__(self, context):
        self.context = context

    def getName(self):
        return self.context.__name__

    def getParent(self):
        return self.context.__parent__

    def isRoot(self):
        return IContainmentRoot.providedBy(self.context)

    def getObjectId(self):
        return u64(self.context._p_oid)

    def getParentState(self):
        if self.getParent() is None:
            return None
        return ZodbObjectStateStub(self.getParent())


class TestZodbInfoViewBreadcrumbs(unittest.TestCase):

    def setUp(self):
        self.root = RootFolderStub()
        self.root._p_oid = p64(1)
        self.foo = PersistentStub()
        self.foo._p_oid = p64(27)
        self.root['foo'] = self.foo
        self.foobar = PersistentStub()
        self.foobar._p_oid = p64(32)
        self.foo['bar'] = self.foobar
        self.unknown = PersistentStub()
        self.unknown._p_oid = p64(15)
        self.unknown_child = PersistentStub()
        self.unknown_child._p_oid = p64(17)
        self.unknown['child'] = self.unknown_child
        self.unparented = PersistentStub()
        self.unparented._p_oid = p64(19)
        self.unparented.__name__ = 'wat'
        self.unnamed = PersistentStub()
        self.unnamed._p_oid = p64(55)
        self.unnamed.__parent__ = self.root

    def createView(self, obj):
        view = ZodbInfoView(obj, TestRequest())
        view.obj = obj
        view.state = ZodbObjectStateStub(view.obj)
        view.requestedTid = None
        view.getRootOid = lambda: 1
        return view

    def test_root(self):
        view = self.createView(self.root)
        self.assertEqual(view.getBreadcrumbs(),
                         [('/', '@@zodbbrowser?oid=0x1'), ])

    def test_non_root(self):
        view = self.createView(self.foo)
        self.assertEqual(view.getBreadcrumbs(),
                         [('/', '@@zodbbrowser?oid=0x1'),
                          ('foo', '@@zodbbrowser?oid=0x1b'), ])

    def test_more_levels(self):
        view = self.createView(self.foobar)
        self.assertEqual(view.getBreadcrumbs(),
                         [('/', '@@zodbbrowser?oid=0x1'),
                          ('foo', '@@zodbbrowser?oid=0x1b'),
                          ('/', None),
                          ('bar', '@@zodbbrowser?oid=0x20'), ])

    def test_unknown(self):
        view = self.createView(self.unknown)
        self.assertEqual(view.getBreadcrumbs(),
                         [('0xf', '@@zodbbrowser?oid=0xf'), ])

    def test_unknown_child(self):
        view = self.createView(self.unknown_child)
        self.assertEqual(view.getBreadcrumbs(),
                         [('0xf', '@@zodbbrowser?oid=0xf'),
                          ('/', None),
                          ('child', '@@zodbbrowser?oid=0x11'), ])

    def test_unparented(self):
        view = self.createView(self.unparented)
        self.assertEqual(view.getBreadcrumbs(),
                         [('/', '@@zodbbrowser?oid=0x1'),
                          ('...', None),
                          ('/', None),
                          ('wat', '@@zodbbrowser?oid=0x13'), ])

    def test_unnamed_direct_child_of_root(self):
        view = self.createView(self.unnamed)
        self.assertEqual(view.getBreadcrumbs(),
                         [('/', '@@zodbbrowser?oid=0x1'),
                          ('???', '@@zodbbrowser?oid=0x37'), ])


class TestZodbInfoView(unittest.TestCase):

    def assertEqual(self, first, second):
        # Align the two values below each other so they're easier to compare.
        if first != second:
            self.fail('\n%r !=\n%r' % (first, second))

    def test_jar_uses_explicit_target_db(self):
        stub_db = DatabaseStub()
        registry = getGlobalSiteManager()
        registry.registerUtility(stub_db, IDatabase, name='<target>')
        self.addCleanup(registry.unregisterUtility,
                        stub_db, IDatabase, name='<target>')
        view = ZodbInfoView(object(), TestRequest())
        self.assertEqual(view.jar.db, stub_db)
        del view
        gc.collect()
        self.assertEqual(stub_db.opened, 0)

    def test_getPath(self):
        view = ZodbInfoView(None, None)
        view.getBreadcrumbs = lambda: [('/', None), ('foo', None),
                                       ('/', None), ('bar baz', None)]
        self.assertEqual(view.getPath(), '/foo/bar baz')

    def test_getBreadcrumbsHTML(self):
        view = ZodbInfoView(None, None)
        view.getBreadcrumbs = lambda: [('/', 'here'),
                                       ('foo>', 'so"there'),
                                       ('/', None),
                                       ('bar<baz', None)]
        self.assertEqual(view.getBreadcrumbsHTML(),
                         '<a href="here">/</a>'
                         '<a href="so&quot;there">foo&gt;</a>'
                         '/bar&lt;baz')

    def test_getDisassembledPickleData(self):
        view = ZodbInfoView(None, None)
        view.state = ZodbObjectStateStub(PersistentStub())
        view.state.pickledState = b''
        self.assertEqual(
            view.getDisassembledPickleData(),
            'ValueError: pickle exhausted before seeing STOP\n\n'
            'ValueError: pickle exhausted before seeing STOP\n'
        )

    def test_listAttributes(self):
        view = ZodbInfoView(None, None)
        view.state = ZodbObjectStateStub(PersistentStub())
        view.state.requestedTid = 42
        view.state.listAttributes = lambda: [('zoinks', 17),
                                             ('scoobysnack', None)]
        self.assertEqual(view.listAttributes(),
                         [ZodbObjectAttribute('scoobysnack', None, 42),
                          ZodbObjectAttribute('zoinks', 17, 42)])

    def test_listAttributes_empty(self):
        view = ZodbInfoView(None, None)
        view.state = ZodbObjectStateStub(PersistentStub())
        view.state.requestedTid = 42
        view.state.listAttributes = lambda: []
        self.assertEqual(view.listAttributes(), [])

    def test_listAttributes_none_exist(self):
        view = ZodbInfoView(None, None)
        view.state = ZodbObjectStateStub(PersistentStub())
        view.state.requestedTid = 42
        view.state.listAttributes = lambda: None
        self.assertEqual(view.listAttributes(), None)

    def test_listItems(self):
        view = ZodbInfoView(None, None)
        view.state = ZodbObjectStateStub(PersistentStub())
        view.state.requestedTid = 42
        view.state.listItems = lambda: [('zoinks', 17),
                                        ('scoobysnack', None)]
        self.assertEqual(view.listItems(),
                         [ZodbObjectAttribute('zoinks', 17, 42),
                          ZodbObjectAttribute('scoobysnack', None, 42)])

    def test_listItems_empty(self):
        view = ZodbInfoView(None, None)
        view.state = ZodbObjectStateStub(PersistentStub())
        view.state.listItems = lambda: []
        self.assertEqual(view.listItems(), [])

    def test_listItems_none_exist(self):
        view = ZodbInfoView(None, None)
        view.state = ZodbObjectStateStub(PersistentStub())
        view.state.listItems = lambda: None
        self.assertEqual(view.listItems(), None)

    def test_loadHistoricalState(self):
        view = ZodbInfoView(None, None)
        view.obj = None
        view.history = [{}]  # injected fault: KeyError('tid')
        self.assertEqual(view._loadHistoricalState(),
                         [{'state': {}, 'error': "KeyError: 'tid'"},
                          {'state': {}, 'error': None}])

    def test_tidToTimestamp(self):
        view = ZodbInfoView(None, None)
        self.assertEqual(view._tidToTimestamp(p64(12345678912345678)),
                         '1905-05-13 03:32:22.050327')
        self.assertEqual(view._tidToTimestamp('something else'),
                         "'something else'")


class HistoryStub(object):
    def __init__(self, tids=()):
        self.tids = list(tids)

    def __len__(self):
        return len(self.tids)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [TransactionRecordStub()
                    for n in range(*idx.indices(len(self)))]
        else:
            return TransactionRecordStub()


class TransactionRecordStub(object):
    tid = p64(0x3c41adbde3708aa)
    status = ' '
    user = b'system /'
    description = b'stuff changed'
    extension = {}

    def __iter__(self):
        return iter(())


class TestZodbHistoryView(RealDatabaseTest):

    def setUp(self):
        RealDatabaseTest.setUp(self)
        self.root = self.conn.root()
        self.root['root'] = RootFolderStub()
        transaction.get().note(u'test setup')
        transaction.get().setUser(u'system')
        transaction.commit()
        provideAdapter(ZodbHistory)
        provideAdapter(getIterableStorage)
        provideAdapter(GenericState)
        provideAdapter(EmptyOOBTreeState)
        provideAdapter(SimpleValueRenderer)

    def _makeView(self, **kw):
        request = TestRequest(**kw)
        view = ZodbHistoryView(self.root, request)
        view.template = lambda: ''
        return view

    def test_render(self):
        view = self._makeView()
        view.render()
        self.assertEqual(view.page_size, 5)
        self.assertEqual(view.page, 0)
        self.assertEqual(view.last_page, 0)
        self.assertEqual(view.last_idx, 2)
        self.assertEqual(view.first_idx, 0)

    def test_render_custom_page(self):
        view = self._makeView(form={'page_size': 2, 'page': 3})
        view.render()
        self.assertEqual(view.page_size, 2)
        self.assertEqual(view.page, 0)
        self.assertEqual(view.last_page, 0)
        self.assertEqual(view.last_idx, 2)
        self.assertEqual(view.first_idx, 0)

    def test_render_bad_tid(self):
        view = self._makeView(form={'tid': '123454321'})
        view.render()
        self.assertEqual(view.page, 0)
        self.assertEqual(view.last_page, 0)
        self.assertEqual(view.last_idx, 2)
        self.assertEqual(view.first_idx, 0)

    def test_getUrl(self):
        view = self._makeView()
        self.assertEqual(view.getUrl(), '@@zodbbrowser_history')
        self.assertEqual(view.getUrl(tid=0x456),
                         '@@zodbbrowser_history?tid=0x456')

    def test_getUrl_preserves_tid(self):
        view = self._makeView(form={'tid': '123'})
        self.assertEqual(view.getUrl(), '@@zodbbrowser_history?tid=123')
        self.assertEqual(view.getUrl(tid=0x456),
                         '@@zodbbrowser_history?tid=0x456')

    def test_findPage(self):
        view = self._makeView()
        view.history = HistoryStub(tids=range(1000, 2000, 10))
        # page 0: [1990, 1980, 1970, 1960, 1950]
        # page 1: [1940, 1930, 1920, 1910, 1900]
        # ...
        # page 18: [1090, 1080, 1070, 1060, 1050]
        # page 19: [1040, 1030, 1020, 1010, 1000]
        self.assertEqual(view.findPage(1990), 0)
        self.assertEqual(view.findPage(1950), 0)
        self.assertEqual(view.findPage(1940), 1)
        self.assertEqual(view.findPage(1050), 18)
        self.assertEqual(view.findPage(1040), 19)
        self.assertEqual(view.findPage(1000), 19)
        self.assertEqual(view.findPage(1337), 0)

    def test_listHistory(self):
        view = self._makeView(form={'tid': '123'})
        view.update()
        prepared_history = view.listHistory()
        self.assertEqual(prepared_history[0]['description'], 'test setup')
        self.assertEqual(prepared_history[0]['user_id'], 'system')
        self.assertEqual(prepared_history[0]['user_location'], '/')

    def test_listHistory_no_transaction_size(self):
        view = self._makeView()
        view.history = HistoryStub(tids=[0x1234])
        view.first_idx = 0
        view.last_idx = 1
        view.page = 0
        prepared_history = view.listHistory()
        self.assertEqual(prepared_history[0]['size'], None)


class TestHelperFunctions(unittest.TestCase):

    def test_getObjectType(self):
        self.assertEqual(getObjectType(NonpersistentStub()),
                         "<class 'zodbbrowser.tests.test_browser.NonpersistentStub'>")
        self.assertEqual(getObjectType(ProxyFactory(NonpersistentStub())),
                         str(Proxy) + " - " +
                         "<class 'zodbbrowser.tests.test_browser.NonpersistentStub'>")

    def test_getObjectTypeShort(self):
        self.assertEqual(getObjectTypeShort(NonpersistentStub()),
                         'NonpersistentStub')
        self.assertEqual(getObjectTypeShort(ProxyFactory(NonpersistentStub())),
                         Proxy.__name__ + ' - NonpersistentStub')


class TestHelperFunctionsWithRealDb(RealDatabaseTest):

    def setUp(self):
        RealDatabaseTest.setUp(self)
        self.root = self.conn.root()
        self.root['root'] = RootFolderStub()
        self.root['root']['item'] = PersistentStub()
        self.root['root']['item']['subitem'] = PersistentStub()
        self.root['detached_item'] = PersistentStub()
        self.root['named_detached_item'] = PersistentStub()
        self.root['named_detached_item'].__name__ = 'named_detached_item'
        transaction.commit()
        provideAdapter(GenericState)

    def test_getObjectPath(self):
        self.assertEqual(
            getObjectPath(self.root['root']['item']['subitem'], None),
            '/item/subitem')

    def test_getObjectPath_no_path_no_name(self):
        oid = u64(self.root['detached_item']._p_oid)
        self.assertEqual(getObjectPath(self.root['detached_item'], None),
                         '0x%x' % oid)

    def test_getObjectPath_no_path_to_root(self):
        self.assertEqual(getObjectPath(self.root['named_detached_item'], None),
                         '/.../named_detached_item')



import unittest
import sys

import transaction
from BTrees.Length import Length
from persistent import Persistent
from persistent.dict import PersistentDict
from zope.app.container.ordered import OrderedContainer
from zope.app.container.sample import SampleContainer
from zope.app.testing import setup
from zope.component import provideAdapter
from zope.container.contained import ContainedProxy
from zope.interface import implementer, alsoProvides, Interface
from zope.interface.interface import InterfaceClass
from zope.interface.verify import verifyObject
from zope.traversing.interfaces import IContainmentRoot

from zodbbrowser.interfaces import IStateInterpreter
from zodbbrowser.history import ZodbObjectHistory
from zodbbrowser.state import (
    ZodbObjectState,
    LoadErrorState,
    GenericState,
    PersistentDictState,
    PersistentMappingState,
    SampleContainerState,
    OrderedContainerState,
    ContainedProxyState,
    FallbackState,
    install_provides_hack,
    uninstall_provides_hack,
    flatten_interfaces,
)
from zodbbrowser.tests.realdb import RealDatabaseTest


class Frob(object):
    pass


@implementer(IContainmentRoot)
class Root(Persistent, SampleContainer):
    pass


class SampleFolder(Persistent, SampleContainer):

    def _newContainerData(self):
        return PersistentDict()


class PersistentObject(Persistent):
    pass # we need a subclass so we get a __dict__


class NamedSampleFolder(SampleFolder):
    __name__ = 'sample_folder'


class SeriouslyBrokenName(Persistent):

    @property
    def __name__(self):
        raise Exception('nobody expects this!')


class IMyInterface(Interface):
    __module__ = 'zodbbrowser.nosuchmodule'


class IMyGoodInterface(Interface):
    pass


class NotAnInterface(object):
    """A stand in for a ZODB Broken object"""
    __module__ = 'zodbbrowser.nosuchmodule'
    __name__ = 'NotAnInterface'


FixedNotAnInterface = InterfaceClass(
    'NotAnInterface', __module__='broken zodbbrowser.nosuchmodule')


class CrashOnUnpickling(object):

    def __getstate__(self):
        return {}

    def __setstate__(self, state):
        raise Exception('oops')


class TestFlattenInterfaces(unittest.TestCase):

    def test(self):
        self.assertEqual(
            flatten_interfaces([
                IMyGoodInterface,
                NotAnInterface,
                (NotAnInterface, IMyGoodInterface)
            ]),
            [
                IMyGoodInterface,
                FixedNotAnInterface,
                FixedNotAnInterface,
                IMyGoodInterface,
            ]
        )


class TestBrokenIntefaces(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(GenericState)
        RealDatabaseTest.setUp(self)
        self.obj = self.conn.root()['obj'] = SampleFolder()
        alsoProvides(self.obj, IMyInterface, IMyGoodInterface)
        sys.modules['zodbbrowser.nosuchmodule'] = sys.modules[__name__]
        transaction.commit()
        sys.modules.pop('zodbbrowser.nosuchmodule', None)
        install_provides_hack()

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()
        sys.modules.pop('zodbbrowser.nosuchmodule', None)
        uninstall_provides_hack()

    def test(self):
        state = ZodbObjectState(self.obj)
        self.assertEqual(state.getError(), None)


class TestZodbObjectState(RealDatabaseTest):

    if not hasattr(unittest.TestCase, 'assertRegex'):
        assertRegex = unittest.TestCase.assertRegexpMatches

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(GenericState)
        provideAdapter(ZodbObjectHistory)
        RealDatabaseTest.setUp(self)
        self.obj = self.conn.root()['obj'] = SampleFolder()
        self.named_obj = self.conn.root()['named_obj'] = NamedSampleFolder()
        self.root = self.conn.root()['application'] = Root()
        self.child = self.root['child'] = SampleFolder()
        transaction.commit()

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def testZodbObjectState(self):
        state = ZodbObjectState(self.obj)
        self.assertEqual(state.getError(), None)
        self.assertEqual(state.listItems(), None)
        self.assertEqual(list(state.listAttributes())[0][0],
                         '_SampleContainer__data')
        self.assertEqual(state.getParent(), None)
        self.assertEqual(state.getParentState(), None)
        self.assertEqual(state.getName(), None)
        self.assertTrue('_SampleContainer__data' in state.asDict().keys())

    def testNameFromClassAttribute(self):
        state = ZodbObjectState(self.named_obj)
        self.assertEqual(state.getName(), 'sample_folder')

    def testRootFolder(self):
        self.assertTrue(ZodbObjectState(self.root).isRoot())
        self.assertFalse(ZodbObjectState(self.obj).isRoot())

    def testParentState(self):
        state = ZodbObjectState(self.child).getParentState()
        self.assertTrue(state.isRoot())

    def testNameResiliency(self):
        obj = self.conn.root()['obj'] = SeriouslyBrokenName()
        transaction.commit()
        state = ZodbObjectState(obj)
        self.assertEqual(state.getName(), None)

    def testUnpickleErrorHandling(self):
        obj = self.conn.root()['obj'] = PersistentObject()
        obj.attribute = CrashOnUnpickling()
        transaction.commit()
        state = ZodbObjectState(obj)
        self.assertEqual(state.getError(), 'Exception: oops')

    def testAdapterLookupErrorHandling(self):
        obj = self.conn.root()['obj'] = Length()
        transaction.commit()
        state = ZodbObjectState(obj)
        self.assertRegex(state.getError(), '^ComponentLookupError: ')


class TestLoadErrorState(unittest.TestCase):

    def setUp(self):
        self.state = LoadErrorState("Failed: because", None)

    def test_interface_compliance(self):
        verifyObject(IStateInterpreter, self.state)

    def test_getError(self):
        self.assertEqual(self.state.getError(), "Failed: because")

    def test_getName(self):
        self.assertEqual(self.state.getName(), None)

    def test_getParent(self):
        self.assertEqual(self.state.getParent(), None)

    def test_listAttributes(self):
        self.assertEqual(self.state.listAttributes(), [])

    def test_listItems(self):
        self.assertEqual(self.state.listItems(), None)

    def test_asDict(self):
        self.assertEqual(self.state.asDict(), {})


class TestGenericState(unittest.TestCase):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(ZodbObjectHistory)

    def tearDown(self):
        setup.placelessTearDown()

    def test_interface_compliance(self):
        verifyObject(IStateInterpreter, GenericState(Frob(), {}, None))

    def test_getError(self):
        self.assertEqual(GenericState(Frob(), {}, None).getError(), None)

    def test_getName_no_name(self):
        self.assertEqual(GenericState(Frob(), {}, None).getName(), None)

    def test_getName_with_name(self):
        state = GenericState(Frob(), {'__name__': 'xyzzy'}, None)
        self.assertEqual(state.getName(), 'xyzzy')

    def test_getParent_no_parent(self):
        self.assertEqual(GenericState(Frob(), {}, None).getParent(), None)

    def test_getParent_with_parent(self):
        parent = Frob()
        state = GenericState(Frob(), {'__parent__': parent}, None)
        self.assertEqual(state.getParent(), parent)

    def test_listAttributes(self):
        state = GenericState(Frob(), {'foo': 1, 'bar': 2, 'baz': 3}, None)
        self.assertEqual(sorted(state.listAttributes()),
                         [('bar', 2), ('baz', 3), ('foo', 1)])

    def test_listItems(self):
        state = GenericState(Frob(), {'foo': 1, 'bar': 2, 'baz': 3}, None)
        self.assertEqual(state.listItems(), None)

    def test_asDict(self):
        state = GenericState(Frob(), {'foo': 1, 'bar': 2, 'baz': 3}, None)
        self.assertEqual(state.asDict(), {'foo': 1, 'bar': 2, 'baz': 3})


class TestGenericStateWithHistory(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(ZodbObjectHistory)
        RealDatabaseTest.setUp(self)
        self.root = self.conn.root()['root'] = Root()
        self.foo = self.root['foo'] = SampleFolder()
        self.bar = self.root['foo']['bar'] = SampleFolder()
        transaction.commit()
        self.foo.__name__ = 'new'
        transaction.commit()

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def test_getParent_no_tid(self):
        state = GenericState(self.bar, {'__parent__': self.foo}, None)
        self.assertEqual(state.getParent().__name__, 'new')

    def test_getParent_old_tid(self):
        self.bar._p_activate()
        tid = self.bar._p_serial
        state = GenericState(self.bar, {'__parent__': self.foo}, tid)
        # previously we mistakenly thought we ought to rewind the parent
        # object to the old state, but that's (a) unnecessary --
        # ZodbObjectState.getParentState() takes care of that -- and (b)
        # very very dangerous: https://launchpad.net/zodbbrowser/+bug/487243
        self.assertEqual(state.getParent().__name__, 'new')


class TestPersistentDictSate(unittest.TestCase):

    def test_listItems(self):
        state = PersistentDictState(Frob(),
                                    {'data': dict(a=42, b=23, c=17)},
                                    None)
        self.assertEqual(state.listItems(),
                         [('a', 42), ('b', 23), ('c', 17)])

    def test_listItems_no_data(self):
        # shouldn't happen, but let's display what exists in the DB instead
        # of crashing
        state = PersistentDictState(Frob(), {}, None)
        self.assertEqual(state.listItems(), [])


class TestPersistentMappingSate(unittest.TestCase):

    def test_listItems(self):
        state = PersistentMappingState(Frob(),
                                       {'data': dict(a=42, b=23, c=17)},
                                       None)
        self.assertEqual(state.listItems(),
                         [('a', 42), ('b', 23), ('c', 17)])

    def test_listItems_no_data(self):
        # shouldn't happen, but let's display what exists in the DB instead
        # of crashing
        state = PersistentMappingState(Frob(), {}, None)
        self.assertEqual(state.listItems(), [])


class TestSampleContainerState(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(ZodbObjectHistory)
        provideAdapter(PersistentDictState)
        provideAdapter(PersistentMappingState)
        RealDatabaseTest.setUp(self)
        self.container = self.conn.root()['container'] = SampleFolder()
        self.container['foo'] = 1
        self.container['bar'] = 2
        transaction.commit()
        self.state = SampleContainerState(None, self.container.__getstate__(),
                                          None)

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def test_listItems(self):
        self.assertEqual(list(self.state.listItems()),
                         [('bar', 2), ('foo', 1)])

    def test_listItems_no_data(self):
        state = SampleContainerState(None, SampleFolder().__getstate__(),
                                     None)
        self.assertEqual(list(state.listItems()), [])


class TestOrderedContainerState(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(ZodbObjectHistory)
        RealDatabaseTest.setUp(self)
        self.container = self.conn.root()['container'] = OrderedContainer()
        self.container['foo'] = 1
        self.container['bar'] = 2
        transaction.commit()
        self.tid = self.container._p_serial
        self.state = OrderedContainerState(self.container,
                                           self.container.__getstate__(),
                                           self.tid)

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def test_listItems(self):
        self.assertEqual(list(self.state.listItems()),
                         [('foo', 1), ('bar', 2)])

    def test_listItems_does_not_change_persistent_objects(self):
        self.container['baz'] = 3
        transaction.commit()
        # State still shows old history
        self.assertEqual(list(self.state.listItems()),
                         [('foo', 1), ('bar', 2)])
        # This doesn't affect any real objects:
        self.assertEqual(list(self.container.items()),
                         [('foo', 1), ('bar', 2), ('baz', 3)])

    def test_listItems_nonpersistent_order(self):
        # I once saw a live OrderedContainer that had a plain mutable
        # builtin list in its _order attribute
        self.container._order = list(self.container._order)
        transaction.commit()
        self.tid = self.container._p_serial
        self.state = OrderedContainerState(self.container,
                                           self.container.__getstate__(),
                                           self.tid)
        self.assertEqual(list(self.state.listItems()),
                         [('foo', 1), ('bar', 2)])


class TestContainedProxyState(unittest.TestCase):

    def setUp(self):
        self.parent = SampleFolder()
        self.frob = Frob()
        self.proxy = ContainedProxy(self.frob)
        self.proxy.__parent__ = self.parent
        self.proxy.__name__ = 'frob'
        self.state = ContainedProxyState(
            self.proxy, self.proxy.__getstate__(), None)

    def test_interface_compliance(self):
        verifyObject(IStateInterpreter, self.state)

    def test_getError(self):
        self.assertEqual(self.state.getError(), None)

    def test_getName(self):
        self.assertEqual(self.state.getName(), 'frob')

    def test_getParent(self):
        self.assertEqual(self.state.getParent(), self.parent)

    def test_listAttributes(self):
        self.assertEqual(
            self.state.listAttributes(),
            [
                ('__name__', 'frob'),
                ('__parent__', self.parent),
                ('proxied_object', self.frob),
            ]
        )

    def test_listItems(self):
        self.assertEqual(self.state.listItems(), [])

    def test_asDict(self):
        self.assertEqual(
            self.state.asDict(),
            {
                '__name__': 'frob',
                '__parent__': self.parent,
                'proxied_object': self.frob,
            }
        )


class TestFallbackState(unittest.TestCase):

    def setUp(self):
        self.state = FallbackState(Frob(), object(), None)

    def test_interface_compliance(self):
        verifyObject(IStateInterpreter, self.state)

    def test_getError(self):
        self.assertEqual(self.state.getError(), None)

    def test_getName(self):
        self.assertEqual(self.state.getName(), None)

    def test_getParent(self):
        self.assertEqual(self.state.getParent(), None)

    def test_listAttributes(self):
        self.assertEqual(self.state.listAttributes(),
                         [('pickled state', self.state.state)])

    def test_listItems(self):
        self.assertEqual(self.state.listItems(), None)

    def test_asDict(self):
        self.assertEqual(self.state.asDict(),
                         {'pickled state': self.state.state})




import logging
import itertools
import collections
import re
from functools import partial

from ZODB.utils import u64, oid_repr
from persistent import Persistent
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from zope.component import adapter
from zope.interface.declarations import ProvidesClass
from zope.interface import implementer, Interface
from zope.security.proxy import removeSecurityProxy


from zodbbrowser.compat import basestring, escape
from zodbbrowser.interfaces import IValueRenderer, IObjectHistory


# Persistent has a __repr__ now that shows the OID, but it's shown poorly
# (as a Python string full of backslash escapes) and includes lots of
# irrelevant details (in-memory object address, repr of the Connection object).
CLASSES_WITH_BAD_REPR = (object, Persistent)
try:
    from BTrees._base import Bucket, Set, Tree, TreeSet
except ImportError:
    pass
else:
    CLASSES_WITH_BAD_REPR += (Bucket, Set, Tree, TreeSet)


log = logging.getLogger(__name__)


MAX_CACHE_SIZE = 1000
TRUNCATIONS = {}
TRUNCATIONS_IN_ORDER = collections.deque()
next_id = partial(next, itertools.count(1))


def resetTruncations(): # for tests only!
    global next_id
    next_id = partial(next, itertools.count(1))
    TRUNCATIONS.clear()
    TRUNCATIONS_IN_ORDER.clear()


def pruneTruncations():
    while len(TRUNCATIONS_IN_ORDER) > MAX_CACHE_SIZE:
        del TRUNCATIONS[TRUNCATIONS_IN_ORDER.popleft()]


def truncate(text):
    id = 'tr%d' % next_id()
    TRUNCATIONS[id] = text
    TRUNCATIONS_IN_ORDER.append(id)
    return id


@adapter(Interface)
@implementer(IValueRenderer)
class GenericValue(object):
    """Default value renderer.

    Uses the object's __repr__, truncating if too long.
    """

    def __init__(self, context):
        self.context = context

    if hasattr(object.__repr__, '__func__'):  # pragma: nocover
        # PyPy
        def _same_method(self, a, b):
            return getattr(a, '__func__', None) is b.__func__
    else:
        # CPython
        def _same_method(self, a, b):
            return a is b

    def _has_no_repr(self, obj):
        obj_repr = getattr(obj.__class__, '__repr__', None)
        return any(self._same_method(obj_repr, cls.__repr__)
                   for cls in CLASSES_WITH_BAD_REPR)

    def _repr(self):
        # hook for subclasses
        if self._has_no_repr(self.context):
            # Special-case objects with the default __repr__ (LP#1087138)
            if isinstance(self.context, Persistent):
                return '<%s.%s with oid %s>' % (
                    self.context.__class__.__module__,
                    self.context.__class__.__name__,
                    oid_repr(self.context._p_oid))
        try:
            return repr(self.context)
        except Exception:
            try:
                return '<unrepresentable %s>' % self.context.__class__.__name__
            except Exception:
                return '<unrepresentable>'

    def render(self, tid=None, can_link=True, limit=200):
        text = self._repr()
        if len(text) > limit:
            id = truncate(text[limit:])
            text = '%s<span id="%s" class="truncated">...</span>' % (
                escape(text[:limit], False), id)
        else:
            text = escape(text, False)
        if not isinstance(self.context, (basestring, bytes)):
            try:
                n = len(self.context)
            except Exception:
                pass
            else:
                if n == 1: # this is a crime against i18n, but oh well
                    text += ' (%d item)' % n
                else:
                    text += ' (%d items)' % n
        return text


def join_with_commas(html, open, close):
    """Helper to join multiple html snippets into a struct."""
    prefix = open + '<span class="struct">'
    suffix = '</span>'
    for n, item in enumerate(html):
        if n == len(html) - 1:
            trailer = close
        else:
            trailer = ','
        if item.endswith(suffix):
            item = item[:-len(suffix)] + trailer + suffix
        else:
            item += trailer
        html[n] = item
    return prefix + '<br />'.join(html) + suffix


@adapter(basestring)
@implementer(IValueRenderer)
class StringValue(GenericValue):
    """String renderer."""

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True, limit=200, threshold=4):
        newline = b'\n' if isinstance(self.context, bytes) else u'\n'
        if self.context.count(newline) <= threshold:
            return GenericValue.render(self, tid, can_link=can_link,
                                       limit=limit)
        else:
            if isinstance(self.context, bytes):
                context = self.context.decode('latin-1').encode('ascii',
                                                            'backslashreplace')
            else:
                context = self.context
            if isinstance(self.context, str):
                prefix = ''
            else:
                prefix = 'u'
            lines = [re.sub(r'^[ \t]+',
                            lambda m: '&nbsp;' * len(m.group(0).expandtabs()),
                            escape(line, False))
                     for line in context.splitlines()]
            nl = '<br />' # hm, maybe '\\n<br />'?
            if sum(map(len, lines)) > limit:
                head = nl.join(lines[:5])
                tail = nl.join(lines[5:])
                id = truncate(tail)
                return (prefix + "'<span class=\"struct\">" + head + nl
                        + '<span id="%s" class="truncated">...</span>' % id
                        + "'</span>")
            else:
                return (prefix + "'<span class=\"struct\">" + nl.join(lines)
                        + "'</span>")


@adapter(tuple)
@implementer(IValueRenderer)
class TupleValue(object):
    """Tuple renderer."""

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True, threshold=100):
        html = []
        for item in self.context:
            html.append(IValueRenderer(item).render(tid, can_link))
        if len(html) == 1:
            html.append('') # (item) -> (item, )
        result = '(%s)' % ', '.join(html)
        if len(result) > threshold or '<span class="struct">' in result:
            if len(html) == 2 and html[1] == '':
                return join_with_commas(html[:1], '(', ', )')
            else:
                return join_with_commas(html, '(', ')')
        return result


@adapter(list)
@implementer(IValueRenderer)
class ListValue(object):
    """List renderer."""

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True, threshold=100):
        html = []
        for item in self.context:
            html.append(IValueRenderer(item).render(tid, can_link))
        result = '[%s]' % ', '.join(html)
        if len(result) > threshold or '<span class="struct">' in result:
            return join_with_commas(html, '[', ']')
        return result


@adapter(dict)
@implementer(IValueRenderer)
class DictValue(object):
    """Dict renderer."""

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True, threshold=100):
        html = []
        for key, value in sorted(self.context.items()):
            html.append(IValueRenderer(key).render(tid, can_link) + ': ' +
                        IValueRenderer(value).render(tid, can_link))
        if (sum(map(len, html)) < threshold and
                '<span class="struct">' not in ''.join(html)):
            return '{%s}' % ', '.join(html)
        else:
            return join_with_commas(html, '{', '}')


@adapter(Persistent)
@implementer(IValueRenderer)
class PersistentValue(object):
    """Persistent object renderer.

    Uses __repr__ and makes it a hyperlink to the actual object.
    """

    view_name = '@@zodbbrowser'
    delegate_to = GenericValue

    def __init__(self, context):
        self.context = removeSecurityProxy(context)

    def render(self, tid=None, can_link=True):
        obj = self.context
        url = '%s?oid=0x%x' % (self.view_name, u64(self.context._p_oid))
        if tid is not None:
            url += "&tid=0x%x" % u64(tid)
            try:
                oldstate = IObjectHistory(self.context).loadState(tid)
                clone = self.context.__class__.__new__(self.context.__class__)
                clone.__setstate__(oldstate)
                clone._p_oid = self.context._p_oid
                obj = clone
            except Exception:
                log.debug('Could not load old state for %s 0x%x',
                          self.context.__class__, u64(self.context._p_oid))
        value = self.delegate_to(obj).render(tid, can_link=False)
        if can_link:
            return '<a class="objlink" href="%s">%s</a>' % (escape(url, True),
                                                            value)
        else:
            return value


@adapter(PersistentMapping)
class PersistentMappingValue(PersistentValue):
    delegate_to = DictValue


@adapter(PersistentList)
class PersistentListValue(PersistentValue):
    delegate_to = ListValue


if PersistentMapping is PersistentDict:
    # ZODB 3.9 deprecated PersistentDict and made it an alias for
    # PersistentMapping.  I don't know a clean way to conditionally disable the
    # <adapter> directive in ZCML to avoid conflicting configuration actions,
    # therefore I'll register a decoy adapter registered for a decoy class.
    # This adapter will never get used.

    class DecoyPersistentDict(PersistentMapping):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""

    @adapter(DecoyPersistentDict)
    class PersistentDictValue(PersistentValue):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""
        delegate_to = DictValue

else:  # pragma: nocover
    @adapter(PersistentDict)
    class PersistentDictValue(PersistentValue):
        delegate_to = DictValue


@adapter(ProvidesClass)
@implementer(IValueRenderer)
class ProvidesValue(GenericValue):
    """zope.interface.Provides object renderer.

    The __repr__ of zope.interface.Provides is decidedly unhelpful.
    """

    def _repr(self):
        return '<Provides: %s>' % ', '.join(i.__identifier__
                                            for i in self.context._Provides__args[1:])




from ZODB.utils import tid_repr
from ZODB.interfaces import IConnection, IStorageIteration
from persistent import Persistent
from zope.proxy import removeAllProxies
from zope.interface import implementer
from zope.component import adapter

from zodbbrowser.interfaces import IObjectHistory, IDatabaseHistory
from zodbbrowser import cache

try:
    from ZODB.mvccadapter import MVCCAdapterInstance
except ImportError:  # pragma: no-cover
    class MVCCAdapterInstance(object):
        """Placeholder so we can register an adapter that will not be used."""


@adapter(Persistent)
@implementer(IObjectHistory)
class ZodbObjectHistory(object):

    def __init__(self, obj):
        self._obj = removeAllProxies(obj)
        self._connection = self._obj._p_jar
        self._storage = self._connection._storage
        self._oid = self._obj._p_oid
        self._history = None
        self._by_tid = {}

    def __len__(self):
        if self._history is None:
            self._load()
        return len(self._history)

    def _load(self):
        """Load history of changes made to a Persistent object.

        Returns a list of dictionaries, from latest revision to the oldest.
        The dicts have various interesting pieces of data, such as:

            tid -- transaction ID (a byte string, usually 8 bytes)
            time -- transaction timestamp (number of seconds since the Unix epoch)
            user_name -- name of the user responsible for the change
            description -- short description (often a URL)

        See the 'history' method of ZODB.interfaces.IStorage.
        """
        size = 999999999999 # "all of it"; ought to be sufficient
        self._history = self._storage.history(self._oid, size=size)
        self._index_by_tid()

    def _index_by_tid(self):
        for record in self._history:
            self._by_tid[record['tid']] = record

    def __getitem__(self, item):
        if self._history is None:
            self._load()
        d = dict(self._history[item])
        if isinstance(d['user_name'], bytes):
            d['user_name'] = d['user_name'].decode('UTF-8', 'replace')
        if isinstance(d['description'], bytes):
            d['description'] = d['description'].decode('UTF-8', 'replace')
        return d

    def lastChange(self, tid=None):
        if self._history is None:
            self._load()
        if tid in self._by_tid:
            # optimization
            return tid
        # sadly ZODB has no API for get revision at or before tid, so
        # we have to find the exact tid
        for record in self._history:
            # we assume records are ordered by tid, newest to oldest
            if tid is None or record['tid'] <= tid:
                return record['tid']
        raise KeyError('%r did not exist in or before transaction %r' %
                       (self._obj, tid_repr(tid)))

    def loadStatePickle(self, tid=None):
        return self._connection._storage.loadSerial(self._obj._p_oid,
                                                    self.lastChange(tid))

    def loadState(self, tid=None):
        return self._connection.oldstate(self._obj, self.lastChange(tid))

    def rollback(self, tid):
        state = self.loadState(tid)
        if state != self.loadState():
            self._obj.__setstate__(state)
            self._obj._p_changed = True


@adapter(IConnection)
@implementer(IDatabaseHistory)
class ZodbHistory(object):

    def __init__(self, connection):
        self._connection = connection
        self._storage = IStorageIteration(connection._storage)
        self._tids = cache.getStorageTids(self._storage)
        self._iterators = []

    @property
    def tids(self):
        return tuple(self._tids) # readonlify

    def __len__(self):
        return len(self._tids)

    def _addcleanup(self, it):
        self._iterators.append(it)
        return it

    def cleanup(self):
        for it in self._iterators:
            it.close()
        self._iterators = []

    def __iter__(self):
        return self._addcleanup(self._storage.iterator())

    def __getitem__(self, index):
        if isinstance(index, slice):
            assert index.step is None or index.step == 1
            tids = self._tids[index]
            if not tids:
                return []
            return self._addcleanup(self._storage.iterator(tids[0], tids[-1]))
        else:
            tid = self._tids[index]
            return next(self._addcleanup(self._storage.iterator(tid, tid)))


@adapter(MVCCAdapterInstance)
@implementer(IStorageIteration)
def getIterableStorage(storage):
    return storage._storage



import unittest
import tempfile
import shutil
import os
import transaction

from ZODB.FileStorage.FileStorage import FileStorage
from ZODB.DB import DB


class RealDatabaseTest(unittest.TestCase):

    open_db = True

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='test-zodbbrowser-')
        self.addCleanup(shutil.rmtree, self.tmpdir)
        self.db_filename = os.path.join(self.tmpdir, 'Data.fs')
        if self.open_db:
            self.storage = FileStorage(self.db_filename)
            self.addCleanup(self.storage.close)
            self.db = DB(self.storage)
            self.addCleanup(self.db.close)
            self.conn = self.db.open()
            self.addCleanup(self.conn.close)

    def tearDown(self):
        transaction.abort()

    def packDatabase(self):
        self.db.pack()



import unittest
import time

import transaction

from zodbbrowser.cache import expired, MINUTES, getStorageTids
from zodbbrowser.tests.realdb import RealDatabaseTest


class TestCache(unittest.TestCase):

    def test_expired(self):
        now = time.time()
        self.assertTrue(expired({}, 5 * MINUTES))
        self.assertFalse(expired({'last_update': now - 1}, 5 * MINUTES))
        self.assertTrue(expired({'last_update': now - 1 - 5 * MINUTES}, 5 * MINUTES))


class TestStorageTids(RealDatabaseTest):

    def setUp(self):
        super(TestStorageTids, self).setUp()
        self.root = self.conn.root()
        self.root['adam'] = None
        transaction.commit()
        self.root['eve'] = None
        transaction.commit()

    def test_getStorageTids_no_cache(self):
        tids = getStorageTids(self.storage)
        self.assertEqual(len(tids), 3)

    def test_getStorageTids_cached(self):
        tids = getStorageTids(self.storage)
        tids_again = getStorageTids(self.storage)
        self.assertEqual(tids, tids_again)

    def test_getStorageTids_notices_new_commits_when_cache_expires(self):
        tids = getStorageTids(self.storage)[:]
        self.root['bob'] = None
        transaction.commit()
        tids_again = getStorageTids(self.storage, cache_for=-1)
        self.assertEqual(len(tids_again), len(tids) + 1)

    def test_getStorageTids_notices_db_packing(self):
        tids = getStorageTids(self.storage)  # prime the cache
        time.sleep(0.001)  # hmm, will this help?
        self.packDatabase()
        tids_again = getStorageTids(self.storage, cache_for=-1)
        self.assertEqual(
            len(tids_again), 1,
            'the impossible happened: my code has a bug?  or your clock is ticking backwards?\n'
            'tids:       {tids!r}\n'
            'tids_again: {tids_again!r}\n'
            'should be:  {should_be!r}\n'.format(
                tids=tids,
                tids_again=tids_again,
                should_be=[t.tid for t in self.storage.iterator()],
            )
        )



from zodbbrowser.compat import basestring, escape
from zodbbrowser.interfaces import IValueRenderer


ADDED = 'added'
REMOVED = 'removed'
CHANGED = 'changed to'


def compareDicts(new, old):
    """Compare two state dictionaries, return dict."""
    diffs = {}
    for key, value in new.items():
        if key not in old:
            diffs[key] = (ADDED, value, None)
        elif old[key] != value:
            diffs[key] = (CHANGED, value, old[key])
    for key, value in old.items():
        if key not in new:
            diffs[key] = (REMOVED, None, value)
    return diffs


def isascii(s):
    """See if the string can be safely converted to unicode."""
    try:
        s.encode('ascii')
    except UnicodeError:
        return False
    else:
        return True


def compareTuples(new, old):
    """Compare two tuples.

    Return (common_prefix, middle_of_old, middle_of_new, common_suffix)
    """
    first = 0
    for oldval, newval in zip(old, new):
        if oldval != newval:
            break
        first += 1
    last = 0
    for oldval, newval in zip(reversed(old[first:]), reversed(new[first:])):
        if oldval != newval:
            break
        last += 1
    return (old[:first],
            old[first:len(old) - last],
            new[first:len(new) - last],
            old[len(old) - last:])


def compareTuplesHTML(new, old, tid=None, indent=''):
    """Compare two tuples, return HTML."""
    html = [indent + '<div class="diff">\n']
    prefix, removed, added, suffix = compareTuples(new, old)
    if len(prefix) > 0:
        html.append(indent + '  <div class="diffitem %s">\n' % 'same')
        if len(prefix) == 1:
            html.append(indent + '    first item kept the same\n')
        else:
            html.append(indent + '    first %d items kept the same\n' % len(prefix))
        html.append(indent + '  </div>\n')
    for oldval in removed:
        html.append(indent + '  <div class="diffitem %s">\n' % REMOVED)
        html.append(indent + '    %s %s\n' % (
            REMOVED, IValueRenderer(oldval).render(tid)))
        html.append(indent + '  </div>\n')
    for newval in added:
        html.append(indent + '  <div class="diffitem %s">\n' % ADDED)
        html.append(indent + '    %s %s\n' % (
            ADDED, IValueRenderer(newval).render(tid)))
        html.append(indent + '  </div>\n')
    if len(suffix) > 0:
        html.append(indent + '  <div class="diffitem %s">\n' % 'same')
        if len(suffix) == 1:
            html.append(indent + '    last item kept the same\n')
        else:
            html.append(indent + '    last %d items kept the same\n' % len(suffix))
        html.append(indent + '  </div>\n')
    html.append(indent + '</div>\n')
    return ''.join(html)


def compareDictsHTML(new, old, tid=None, indent=''):
    """Compare two state dictionaries, return HTML."""
    html = [indent + '<div class="diff">\n']
    diff = compareDicts(new, old)
    for key, (action, newvalue, oldvalue) in sorted(
            diff.items(), key=lambda kv: (str(type(kv[0])), kv[0])):
        what = action.split()[0]
        html.append(indent + '  <div class="diffitem %s">\n'
                    % escape(what, True))
        if isinstance(key, basestring) and isascii(key):
            html.append(indent + '    <strong>%s</strong>: '
                        % escape(key, False))
        else:
            html.append(indent + '    <strong>%s</strong>: '
                        % IValueRenderer(key).render(tid))
        if (action == CHANGED and isinstance(oldvalue, dict) and
                isinstance(newvalue, dict)):
            html.append('dictionary changed:\n')
            html.append(compareDictsHTML(newvalue, oldvalue, tid,
                                         indent=indent + '    '))
        elif (action == CHANGED and isinstance(oldvalue, tuple) and
              isinstance(newvalue, tuple)):
            html.append('tuple changed:\n')
            html.append(compareTuplesHTML(newvalue, oldvalue, tid,
                                          indent=indent + '    '))
        else:
            html.append(action)
            html.append(' ')
            if action == REMOVED:
                value = oldvalue
            else:
                value = newvalue
            html.append(IValueRenderer(value).render(tid))
            html.append('\n')
        html.append(indent + '  </div>\n')
    html.append(indent + '</div>\n')
    return ''.join(html)




from zope.component import adapter
from zope.interface import implementer

from zodbbrowser.interfaces import IValueRenderer


@adapter(None)
@implementer(IValueRenderer)
class SimpleValueRenderer(object):

    def __init__(self, context):
        self.context = context

    def render(self, tid=None):
        s = repr(self.context)
        if tid:
            s += ' [tid=%s]' % tid
        return s




import errno
import os
import shutil
import socket
import tempfile
import unittest
import sys

import mock
from zope.app.server.servertype import IServerType
from zope.app.testing import setup
from zope.component import provideUtility
from zope.interface import implementer
from ZEO.Exceptions import ClientDisconnected

from zodbbrowser.compat import StringIO
from zodbbrowser.standalone import (
    Options, parse_args, start_server, serve_forever, main, open_db,
    stop_serving)
from zodbbrowser.tests.realdb import RealDatabaseTest


class SocketMixin(object):

    def make_socket(self):
        tempdir = tempfile.mkdtemp(prefix='test-zodbbrowser-')
        self.addCleanup(shutil.rmtree, tempdir)
        sockfilename = os.path.join(tempdir, 'zeosock')
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.addCleanup(sock.close)
        sock.bind(sockfilename)
        sock.listen(0)
        return sockfilename


class TestParseArgs(SocketMixin, unittest.TestCase):

    def test_sys_argv(self):
        with mock.patch('sys.argv', ['zodbbrowser', '/dev/null']):
            options = parse_args()
        self.assertEqual(options.db_filename, '/dev/null')

    def test_listen_port(self):
        options = parse_args(['--listen', '8000', '/dev/null'])
        self.assertEqual(options.listen_on, ('localhost', 8000))

    def test_listen_host_and_port(self):
        options = parse_args(['--listen', '0.0.0.0:8000', '/dev/null'])
        self.assertEqual(options.listen_on, ('0.0.0.0', 8000))

    def test_listen_bad_port(self):
        with self.assertRaises(SystemExit), mock.patch('sys.stderr'):
            parse_args(['--listen', 'xyzzy', '/dev/null'])

    def test_too_many_args(self):
        with self.assertRaises(SystemExit), mock.patch('sys.stderr'):
            parse_args(['Data-1.fs', 'Data-2.fs'])

    def test_too_few_args(self):
        with self.assertRaises(SystemExit), mock.patch('sys.stderr'):
            parse_args([])

    def test_zeo_and_file(self):
        with self.assertRaises(SystemExit), mock.patch('sys.stderr'):
            parse_args(['--zeo', '/tmp/zeosock', 'Data.fs'])

    def test_zeo_storage_no_zeo(self):
        with self.assertRaises(SystemExit), mock.patch('sys.stderr'):
            parse_args(['--storage', '2', 'Data.fs'])

    def test_zeo_port(self):
        options = parse_args(['--zeo', 'localhost:7099'])
        self.assertEqual(options.zeo_address, ('localhost', 7099))

    def test_zeo_bad_port(self):
        with self.assertRaises(SystemExit), mock.patch('sys.stderr'):
            parse_args(['--zeo', 'localhost:xyzzy'])

    def test_zeo_default_port(self):
        options = parse_args(['--zeo', 'localhost'])
        self.assertEqual(options.zeo_address, ('localhost', 8100))

    def test_zeo_storage(self):
        options = parse_args(['--zeo', 'localhost', '--storage', '2'])
        self.assertEqual(options.zeo_storage, '2')

    @unittest.skipUnless(hasattr(socket, 'AF_UNIX'),
                         "No UNIX domain sockets on this platform")
    def test_zeo_socket(self):
        sockfilename = self.make_socket()
        options = parse_args(['--zeo', sockfilename])
        self.assertEqual(options.zeo_address, sockfilename)

    def test_bad_zeo_socket(self):
        with self.assertRaises(SystemExit), mock.patch('sys.stderr'):
            parse_args(['--zeo', __file__])


class TestOpenDb(unittest.TestCase):

    @unittest.skipIf(sys.platform == 'win32', "This test hangs on Windows")
    def test_zeo(self):
        options = Options()
        options.zeo_address = '/no/such/zeo/socket'
        options.zeo_timeout = 0.001
        with self.assertRaises(ClientDisconnected):
            open_db(options)


@implementer(IServerType)
class FakeServerType(object):
    def create(self, **kw):
        if kw.get('port') == 80:
            raise socket.error(errno.EADDRINUSE, "port busy")
        if kw.get('port') == -1:
            raise socket.error(errno.EINVAL, "bad port")
        sock = mock.Mock()
        sock.getsockname.return_value = ('localhost', 8033)
        return mock.Mock(socket=sock)


class TestStartServer(unittest.TestCase):

    def setUp(self):
        setup.placelessSetUp()
        provideUtility(FakeServerType(), name='FAKE')
        self.options = Options()
        self.options.server_type = 'FAKE'
        self.options.verbose = False
        self.db = None

    def tearDown(self):
        stop_serving()
        setup.placelessTearDown()

    def test_prints_clickable_url(self):
        self.options.verbose = True
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            start_server(self.options, self.db)
        self.assertEqual(mock_stdout.getvalue(),
                         "Listening on http://localhost:8033/\n")

    def test_socket_error_handling_eaddrinuse(self):
        self.options.listen_on = ('localhost', 80)
        with self.assertRaises(SystemExit) as e:
            start_server(self.options, self.db)
        self.assertTrue(
            str(e.exception).startswith("Cannot listen on localhost:80"),
            str(e.exception))

    def test_socket_error_handling_other_kind_of_error(self):
        self.options.listen_on = ('localhost', -1)
        with self.assertRaises(socket.error):
            start_server(self.options, self.db)


class TestServeForever(unittest.TestCase):

    @mock.patch('asyncore.socket_map', {42: None})
    @mock.patch('asyncore.poll', mock.Mock(side_effect=KeyboardInterrupt))
    def test(self):
        serve_forever()


class TestMain(RealDatabaseTest):

    open_db = False

    def setUp(self):
        setup.placelessSetUp()
        RealDatabaseTest.setUp(self)

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    @mock.patch('logging.basicConfig', mock.Mock())
    @mock.patch('zodbbrowser.standalone.open_db', mock.Mock())
    @mock.patch('zodbbrowser.standalone.configure', mock.Mock())
    @mock.patch('zodbbrowser.standalone.start_server', mock.Mock())
    @mock.patch('zodbbrowser.standalone.serve_forever', mock.Mock())
    def test(self):
        main(['--quiet', '--listen', '0', '--rw', self.db_filename])



#!/usr/bin/env python
import io
import os
import re
from setuptools import setup, find_packages


def get_version_and_homepage():
    d = {}
    r = re.compile('''^(__[a-z]+__) = ["'](.+)["']$''')
    for line in read_file('src', 'zodbbrowser', '__init__.py').splitlines():
        m = r.match(line)
        if m:
            d[m.group(1)] = m.group(2)
    return d['__version__'], d['__homepage__']


def read_file(*relative_filename):
    here = os.path.dirname(__file__)
    with io.open(os.path.join(here, *relative_filename), encoding='UTF-8') as f:
        return f.read()


def linkify_bugs(text):
    text = re.sub(r'\bLP#(\d+)\b',
                  r'`\g<0> <https://pad.lv/\1>`__', text)
    text = re.sub(r'\bGH #(\d+)\b',
                  r'`\g<0> <https://github.com/mgedmin/zodbbrowser/issues/\1>`__', text)
    return text


def get_long_description():
    return linkify_bugs(
        read_file('README.rst') +
        '\n\n' +
        read_file('CHANGES.rst')
    )


version, homepage = get_version_and_homepage()
long_description = get_long_description()

setup(
    name="zodbbrowser",
    license='ZPL 2.1',
    author="Marius Gedminas",
    author_email="marius@pov.lt",
    maintainer="Programmers of Vilnius",
    maintainer_email="marius@pov.lt",
    description="ZODB browser",
    long_description=long_description,
    version=version,
    url=homepage,
    keywords='ZODB database interactive history browser',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    packages=find_packages('src'),
    include_package_data=True,
    zip_safe=False,
    package_dir={'': 'src'},
    install_requires=[
        "BTrees",
        "ZODB",
        "ZEO",
        "persistent",
        "transaction",
        "zope.app.pagetemplate",
        "zope.app.publication",
        "zope.component",
        "zope.interface",
        "zope.location",
        "zope.publisher",
        "zope.security",
        "zope.traversing",
        "zope.cachedescriptors",
        # dependencies just for the standalone app
        "zope.app.authentication",
        "zope.app.component",
        "zope.securitypolicy",
        "zope.app.server",
        "zope.app.session", # purely BBB for old Data.fs'es
        "zope.app.zcmlfiles",
        "zope.server",
        "zope.error",
        "zope.exceptions",
        "zope.session",
        # dependencies that easy_install pulls in via setuptools extras
        # but that are dropped when you do pip install, *sigh*
        "zope.hookable",
        "RestrictedPython",
        ],
    extras_require=dict(
        test=[
            'mock',
            "zope.app.folder",
            "zope.app.container",
            "zope.app.testing",
            "zope.testbrowser >= 5.1",
            "lxml",
            "cssselect",
        ],
        # extras for backwards compatibility only, not to break
        # people's setups if they depend on zodbbrowser[app]
        app=[],
    ),
    entry_points={
        'console_scripts': [
            'zodbbrowser = zodbbrowser.standalone:main',
        ],
        'z3c.autoinclude.plugin': [
            'target = plone',
        ],
    },
)
